//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis                         
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-                                
//*  http://apps.jcns.fz-juelich.de/frida                                   
//**************************************************************************************************

//! \file  expr.cpp
//! \brief Parse expressions using CRef, CContext, CResult, virtual CNode

#include <string>
#include <vector>

#include <trivia/vector_ops.hpp>

#include "defs.hpp"
#include "olf.hpp"
#include "mem.hpp"
#include "slice.hpp"
#include "expr.hpp"


//**************************************************************************************************
//*  CContext                                                               
//**************************************************************************************************


//! Constructor. Sets k,j,i to arguments. Sets other members to defaults.

CContext::CContext( int _k, int _j, int _i )
    : k(_k), j(_j), i(_i),                     // set from arguments
      vt(0), dim(_1), want_error(false), nv(0) // can be changed later
{
    if ( k==(int)-1 )
        return; // not sure whether k=-1 makes sense
    POlo f = NOlm::mem_get(k);
    if ( j==(int)-1 ) {
        if ( f->nJ()==1 )
            j = 0;
    } else if ( j>=f->nJ() )
        throw string("invalid evaluation context, j=")+S(j)+">=nJ="+S(f->nJ());

    POld fd = P2D(f);
    if ( !fd ) {
        if ( i!=(int)-1 )
            throw string("invalid evaluation context, i set for curve file");
        return;
    }

    if ( j==(int)-1 )
        return;
    int ni = fd->nPts(j);
    if ( i==(int)-1 ) {
        if ( ni==1 )
            i = 0;
    } else if ( i>=ni )
        throw string("invalid evaluation context for j=")+S(j)+": i="+
            S(i)+">=ni="+S(ni);        
}


//! Sets request of a vector of doubles, corresponding to input points 't'.

void CContext::request_VT( const vector<double> *_vt )
{
    dim = _VT;
    nv = _vt->size();
    vt = _vt;
    if( !triv::is_equidist( &vt_step, *vt ) || vt_step<0 )
        vt_step = 0;
}

//! Sets request of a vector of _nv indices.

void CContext::request_VI( int _nv )
{
    dim = _VI;
    nv = _nv;
}

//! Returns textual representation of the class instance.

string CContext::context_info() const
{
    string out = "context: ";
    if      ( dim==_1 )
        out += "dim 1, ";
    else if ( dim==_VI )
        out += "dimVI, ";
    else if ( dim==_VT )
        out += "dimVT, ";
    out += "k = " + S(k) + ", ";
    out += "j = " + S(j) + ", ";
    out += "i = " + S(i) + ", ";
    out += "nv = " + S(nv) + ", ";
    if( vt )
        out += "size(vt) = " + S(vt->size()) + ", ";
    else
        out += "no vt, ";
    out += "want_error = " + S(want_error);
    return out;
}


//**************************************************************************************************
//*  CResult                                                                
//**************************************************************************************************


//! Returns textual representation of the class instance, for debugging.

string CResult::result_info() const
{
    string out = "result: ";
    out += "vectorial = " + S(vectorial) + ", ";
    out += "r = " + S(r) + ", ";
    out += "size(v) = " + S(v.size());
    return out;
}


//**************************************************************************************************
//*  CRef (references like y[k,j,0] or p3[2] or ...)                        
//**************************************************************************************************


//! Returns file index k in given context.

int CRef::get_k( const CContext& ctx ) const
{
    int k;
    if ( tk ) {
        CResult tmp;
        tk->tree_val( tmp, ctx );
        if ( tmp.vectorial )
            throw string("BUG: K must not depend on i");
        k = (int)( tmp.r + 0.5 );
    } else {
        k = ctx.k;
    }
    if ( k==(int)-1 )
        throw string("missing or invalid k");
    NOlm::mem_get( k ); // to assert existence of file
    return k;
}


//! Returns spectrum index j in given context.

int CRef::get_j( const CContext& ctx, int nj ) const
{
    try {
        int j;
        if ( tj ) {
            CResult tmp;
            tj->tree_val( tmp, ctx );
            if ( tmp.vectorial ) {
                throw string("BUG: unexpected return type");
            } else
                j = (int)( tmp.r + 0.5 );
        } else {
            j = ctx.j;
        }
        if ( j==-1 )
            throw string("index j not defined");
        else if ( j>=nj )
            throw string("index j=" + S(j) + " out of bounds");
        return j;
    } catch( string& ex ) {
        throw ex + " in Ref::get_j " + ref_info() +
            " {context=" + ctx.context_info() + "}";
    }
}

//! Returns string representation of the class instance.

string CRef::ref_info() const
{
    string out = "ref: ";
    if ( tk )
        out += "tk[" + tk->tree_info() + "], ";
    else
        out += "tk=0, ";
    if ( tj )
        out += "tj[" + tj->tree_info() + "], ";
    else
        out += "tj=0, ";
    if ( ti )
        out += "ti[" + ti->tree_info() + "]";
    else
        out += "ti=0";
    return out;
}

//**************************************************************************************************
//*  CNode: status requests                                                 
//**************************************************************************************************


//! Returns number of curve parameters used in node and below.

int CNode::npar() const
{
    int np=0;
    npar_exec( &np ); // implemented with pointer argument to allow recursion
    return np;
}


//! Does the node depend on k?

bool CNode::k_dependent() const
{
    bool kd=false;
    kdep_exec( &kd );
    return kd;
}


//**************************************************************************************************
//*  CNode: low-level tree evaluation                                       
//**************************************************************************************************


//! Evaluates the node, and sets *ret to the result, rounded for use as index.

void CNode::tree_uival( int *ret, const CContext& ctx ) const
{
    CResult val;
    tree_val( val, ctx );
    // since we have no integer data type, convert real by rounding
    int i = (int) round( val.r );
    if( i<0 )
        throw string("found negative index ") + S(i);
    *ret = (int) i;
}


//**************************************************************************************************
//*  CNode: public interface to tree evaluation                             
//**************************************************************************************************


//! Evaluates the node, and sets *ret and *dret according to the result.

void CNode::tree_point_val(
    double *ret, double *dret, int k, int j, int i ) const
{
    CContext ctx( k, j, i );
    ctx.want_error = (bool)dret;
    CResult val;
    tree_val( val, ctx );
    if ( val.vectorial )
        throw string("BUG: tree_point_val did not get scalar result");
    *ret = val.r;
    if( dret )
        *dret = val.dr;
}

//! Evaluates the node, and sets vectors *ret and *dret according to the result.

void CNode::tree_vec_val(
    vector<double> *ret, vector<double> *dret, int k, int j ) const
{
    CResult val;
    CContext ctx( k, j );
    int nret = ret->size();
    ctx.request_VI( nret );
    ctx.want_error = (bool)dret;
    tree_val( val, ctx );
    if ( val.vectorial ){
        *ret = val.v;
        if( dret )
            *dret = val.dv;
    } else {
        ret->assign( nret, val.r );
        if( dret && val.dr )
            dret->assign( nret, val.dr );
    }
}