//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis                         
//*  (C) Joachim Wuttke 2001-                                               
//*  http://apps.jcns.fz-juelich.de/frida                                   
//**************************************************************************************************

//! \file  geni.cpp
//! \brief collection NGeni of generalized integrals

#include <iostream>
#include <vector>
#include <map>
#include <boost/format.hpp>
using boost::format;

#include <trivia/string_ops.hpp>
#include "defs.hpp"
#include "coord.hpp"
#include "expr.hpp"
#include "geni.hpp"
#include "xax_lex.hpp"

//**************************************************************************************************
//*   Low-level implementation
//**************************************************************************************************

//! Compute minimum value of a vector.

void geni_valmin( double* r, double* dr, int n, const vector<PResult>& a )
{
    int idx = 0;
    *r = a[0]->v[0];
    for ( int i=1; i<n; ++i ) {
        if ( a[0]->v[i]<*r ) {
            *r = a[0]->v[i];
            idx = i;
        }
    }
    if( dr )
        *dr = a[0]->dv[idx];
}

//! Compute maximum value of a vector.

void geni_valmax( double* r, double* dr, int n, const vector<PResult>& a )
{
    int idx = 0;
    *r = a[0]->v[0];
    for ( int i=1; i<n; ++i ) {
        if ( a[0]->v[i]>*r ) {
            *r = a[0]->v[i];
            idx = i;
        }
    }
    if( dr )
        *dr = a[0]->dv[idx];
}

//! Return x[i] with i for which y[i] is minimal.

void geni_argmin( double* r, double* dr, int n, const vector<PResult>& a )
{
    int idx = 0;
    double extr = a[1]->v[0];
    for ( int i=1; i<n; ++i ) {
        if ( a[1]->v[i]<extr ) {
            extr = a[1]->v[i];
            idx = i;
        }
    }
    *r = a[0]->v[idx];
    if( dr )
        *dr = 0; // TODO better estimate
}

//! Return x[i] with i for which y[i] is maximal.

void geni_argmax( double* r, double* dr, int n, const vector<PResult>& a )
{
    int idx = 0;
    double extr = a[1]->v[0];
    for ( int i=1; i<n; ++i ) {
        if ( a[1]->v[i]>extr ) {
            extr = a[1]->v[i];
            idx = i;
        }
    }
    *r = a[0]->v[idx];
    if( dr )
        *dr = 0; // TODO better estimate
}

//! Return i for which x[i] is minimal.

void geni_idxmin( double* r, double* dr, int n, const vector<PResult>& a )
{
    int idx = 0;
    double extr = a[0]->v[0];
    for ( int i=1; i<n; ++i ) {
        if ( a[0]->v[i]<extr ) {
            extr = a[0]->v[i];
            idx = i;
        }
    }
    *r = idx;
    if( dr )
        *dr = 0; // TODO better estimate
}

//! Return i for which x[i] is maximal.

void geni_idxmax( double* r, double* dr, int n, const vector<PResult>& a )
{
    int idx = 0;
    double extr = a[0]->v[0];
    for ( int i=1; i<n; ++i ) {
        if ( a[0]->v[i]>extr ) {
            extr = a[0]->v[i];
            idx = i;
        }
    }
    *r = idx;
    if( dr )
        *dr = 0; // TODO better estimate
}

//! Sum.

void geni_sum( double* r, double* dr, int n, const vector<PResult>& a )
{
    *r = 0;
    for ( int i=0; i<n; i++ )
        *r += a[0]->v[i];
    if( dr )
        *dr = 0; // TODO better estimate
}

//! Average.

void geni_avge( double* r, double* dr, int n, const vector<PResult>& a )
{
    *r = 0;
    if ( !n )
        throw string("Cannot compute average of empty vector");
    for ( int i=0; i<n; i++ )
        *r += a[0]->v[i];
    *r /= n;
    if( dr )
        *dr = 0; // TODO better estimate
}

//! Standard deviation.

void geni_stdv( double* r, double* dr, int n, const vector<PResult>& a )
{
    if ( n<2 )
        throw string("Cannot compute corrected standard deviation for"
            " sample size < 2");
    double p=0, pp=0, q=0;
    for ( int i=0; i<n; i++ ) {
        pp = p + ( a[0]->v[i] - p ) / i;
        q += ( a[0]->v[i] - p ) * ( a[0]->v[i] - pp );
        p = pp;
    }
    *r = q / (n-1);
    if( dr )
        *dr = 0; // TODO better estimate
}

//! Integral.

void geni_integral( double* r, double* dr, int n, const vector<PResult>& a )
{
    *r = 0;
    for ( int i=0; i<n-1; i++ ) {
        if (a[0]->v[i+1]<=a[0]->v[i])
            throw "a not sorted";
        *r += (a[0]->v[i+1]-a[0]->v[i])*(a[1]->v[i+1]+a[1]->v[i])/2;
    }
    if( dr )
        *dr = 0; // TODO better estimate
}

//! Center of gravity in x, weighed by y.

void geni_cog( double* r, double* dr, int n, const vector<PResult>& a )
{
    double r0 = 0, r1 = 0;
    for ( int i=0; i<n; i++ ) {
        r0 += a[1]->v[i];
        r1 += a[1]->v[i] * a[0]->v[i];
    }
    *r = r0>0 ? r1/r0 : 0;
    if( dr )
        *dr = 0; // TODO better estimate
}

//! Standard deviation in x, weighed by y.

void geni_width( double* r, double* dr, int n, const vector<PResult>& a )
{
    double r0 = 0, r1 = 0, r2 = 0;
    for ( int i=0; i<n; i++ ) {
        r0 += a[1]->v[i];
        r1 += a[1]->v[i] * a[0]->v[i];
    }
    if( r0<=0 )
        throw "r0 <= 0";
    double center = r1/r0;
    for ( int i=0; i<n; i++ ) {
        r2 += a[1]->v[i] * SQR(a[0]->v[i]-center);
    }
    *r = sqrt( r2/r0 );
    if( dr )
        *dr = 0; // TODO better estimate
}

//! Correlation coefficient.
void geni_corr( double* r, double* dr, int n, const vector<PResult>& a )
{
    double r0 = 0, r1 = 0, s0 = 0, s1 = 0, v = 0;
    for ( int i=0; i<n; i++ ) {
        r0 += a[0]->v[i] / n;
        r1 += a[1]->v[i] / n;
    }
    for ( int i=0; i<n; i++ ) {
        s0 += SQR(a[0]->v[i]-r0) / n;
        s1 += SQR(a[1]->v[i]-r1) / n;
        v  += (a[0]->v[i]-r0) * (a[1]->v[i]-r1) / n;
    }
    *r = v / sqrt(s0*s1);
    if( dr )
        *dr = 0; // TODO better estimate
}


//**************************************************************************************************
//*   Registration and retrieval
//**************************************************************************************************

namespace NGeni { // internals:
    map<string,const CGeni*> gmap; // unsorted hash, for expression evaluation
    vector<const CGeni*> glist;    // sorted array, for help text
}

CGeni::CGeni( string _name, int _narg, string _def,
              geni_eval _eval, string _com )
    : name(_name), narg(_narg), eval(_eval), com(_com)
{
    vector<string> def_expr;
    triv::split( _def, def_expr, "," );
    if ( def_expr.size()!=narg )
        throw "BUG: invalid default "+_def+" for CGeni "+name;
    def_arg.resize(narg);
    for ( int iarg=0; iarg<narg; ++iarg )
        def_arg[iarg] = user_xaxparse( def_expr[iarg].c_str() );
}

const CGeni* NGeni::find( string key )
{
    auto pos = gmap.find(key);
    return pos == gmap.end() ? NULL : pos->second;
}

void CGeni::register_me() const
{
    NGeni::gmap.insert( make_pair( name, this ) );
    NGeni::glist.push_back( this );
}

void NGeni::display_genis()
{
    cout << "Generalized integrals:\n";
    for( auto g = glist.begin(); g!=glist.end(); ++g ) {
        string s1 = (*g)->name + "(a";
        if( (*g)->narg>1 )
            s1 += ",b";
        s1 += ")";
        cout << " " << str( format("%-14s") % s1 ) << (*g)->com << "\n";
    }
}

void NGeni::initialize(void)
{
    // operators by precedence (as in xax_yacc.ypp):
    (new CGeni( "valmin", 1, "y", geni_valmin,
                "minimum value of a" ))->register_me();
    (new CGeni( "valmax", 1, "y", geni_valmax,
                "maximum value of a" ))->register_me();
    (new CGeni( "argmin", 2, "x,y", geni_argmin,
                "value of a for which b is minimal" ))->register_me();
    (new CGeni( "argmax", 2, "x,y", geni_argmax,
                "value of a for which b is maximal" ))->register_me();
    (new CGeni( "idxmin", 1, "y", geni_idxmin,
                "index for which a is minimal" ))->register_me();
    (new CGeni( "idxmax", 1, "y", geni_idxmax,
                "index for which a is maximal" ))->register_me();
    (new CGeni( "sum", 1, "y", geni_sum,
                "sum of a" ))->register_me();
    (new CGeni( "avge", 1, "y", geni_avge,
                "average of a" ))->register_me();
    (new CGeni( "stdv", 1, "y", geni_stdv,
                "corrected sample standard deviation of a" ))->register_me();
    (new CGeni( "integral", 2, "x,y", geni_integral,
                "integral d(a) of b" ))->register_me();
    (new CGeni( "cog", 2, "x,y", geni_cog,
                "center of gravity of a weighed with b" ))->register_me();
    (new CGeni( "width", 2, "x,y", geni_width,
                "standard variation in a weighed with b" ))->register_me();
    (new CGeni( "corr", 2, "x,y", geni_corr,
                "correlation coefficient of a and b" ))->register_me();
}