//**************************************************************************************************
//* FRIDA: fast reliable interactive data analysis            
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                   
//* http://apps.jcns.fz-juelich.de/frida                      
//**************************************************************************************************

//! \file  node.cpp
//! \brief Nodes within expression trees, inheriting from virtual base CNode.


#include <algorithm>
#include <iostream>

#include "defs.hpp"
#include "olf.hpp"
#include "mem.hpp"
#include "slice.hpp"
#include "reg.hpp"
#include "func.hpp"
#include "geni.hpp"
#include "integrate.hpp"
#include "var.hpp"
#include "expr.hpp"
#include "node.hpp"
#include "curve.hpp"

extern bool allow_slow_conv; // set in curve.cpp

#define DEBUG 0


//**************************************************************************************************
//* CNodeFun: function/operator node                          
//**************************************************************************************************

CNodeFun::CNodeFun( const class CFunc* _fun, PNode _a0 )
    : fun(_fun), narg(1)
{
    if ( fun->narg!=1 )
        throw "syntax error: function '" + fun->txt + "' needs " +
            S(fun->narg) + " arguments, not 1";
    arg[0] = _a0;
}
CNodeFun::CNodeFun( const class CFunc* _fun, PNode _a0, PNode _a1 )
    : fun(_fun), narg(2)
{
    if ( fun->narg!=2 )
        throw "syntax error: function '" + fun->txt + "' needs " +
            S(fun->narg) + " arguments, not 2";
    arg[0] = _a0;
    arg[1] = _a1;
}
CNodeFun::CNodeFun( const class CFunc* _fun, PNode _a0, PNode _a1, PNode _a2 )
    : fun(_fun), narg(3)
{
    if ( fun->narg!=3 )
        throw "syntax error: function '" + fun->txt + "' needs " +
            S(fun->narg) + " arguments, not 3";
    arg[0] = _a0;
    arg[1] = _a1;
    arg[2] = _a2;
}

//! A function of 1,2,.. arguments (including operators like +,-,.. ).

void CNodeFun::tree_val( CResult& ret, const CContext& ctx ) const
{
    try {
        // evaluate arguments,
        // determine if all results are scalar arguments,
        // determine if any result possess an error
        CResult a[maxarg];
        bool is_scalar = true;
        bool err_input = false;
        for ( int iarg=0; iarg<narg; ++iarg ) {
            arg[iarg]->tree_val( a[iarg], ctx );
            if ( a[iarg].vectorial )
                is_scalar = false;
            if ( a[iarg].has_err() )
                err_input = true;
        }
        // determine size of vectorial arguments
        int n=0;
        if( !is_scalar ) {
            for ( int iarg=0; iarg<narg; ++iarg ) {
                if ( !a[iarg].vectorial )
                    continue;
                if ( n==0 )
                    n = a[iarg].v.size();
                else if ( a[iarg].v.size() != n )
                    throw string("vector arguments have different size");
            }
            ret.preset_v( n, ctx.want_error );
        }
        // now evaluate the function
        if       ( const CFunc1* fu = dynamic_cast<const CFunc1*>( fun ) ) {
            if( ctx.want_error && fu->d1 && err_input ) {
                if ( is_scalar ) {
                    (*(fu->d1))( ret.r, ret.dr, a[0].r, a[0].dr );
                } else {
                    for ( int i=0; i<n; ++i )
                        (*(fu->d1))( ret.v[i], ret.dv[i], a[0].v[i], a[0].dv[i] );
                }
            } else {
                if ( is_scalar )
                    ret.set_r( (*(fu->f1))( a[0].r ) );
                else
                    for ( int i=0; i<n; ++i )
                        ret.v[i] = (*(fu->f1))( a[0].v[i] );
            }
        } else if ( const CFunc2* fu = dynamic_cast<const CFunc2*>( fun ) ) {
            if( ctx.want_error && ( ( fu->d2 && err_input ) ||
                                    fu->txt=="+-" ) ) {
                if ( is_scalar ) {
                    (*(fu->d2))( ret.r, ret.dr, a[0].r, a[0].dr, a[1].r, a[1].dr );
                } else {
                    for ( int i=0; i<n; ++i )
                        (*(fu->d2))( ret.v[i], ret.dv[i],
                                      a[0].to_r(i), a[0].to_dr(i),
                                      a[1].to_r(i), a[1].to_dr(i) );
                }
            } else {
                if       ( !a[0].vectorial && !a[1].vectorial )
                    ret.set_r( (*(fu->f2))( a[0].r, a[1].r ) );
                else if  ( !a[0].vectorial && a[1].vectorial ) {
                    for ( int i=0; i<n; ++i )
                        ret.v[i] = (*(fu->f2))( a[0].r, a[1].v[i] );
                } else if ( a[0].vectorial && !a[1].vectorial ) {
                    for ( int i=0; i<n; ++i )
                        ret.v[i] = (*(fu->f2))( a[0].v[i], a[1].r );
                } else if ( a[0].vectorial && a[1].vectorial ) {
                    for ( int i=0; i<n; ++i )
                        ret.v[i] = (*(fu->f2))( a[0].v[i], a[1].v[i] );
                }
            }
        } else if ( const CFunc3* fu = dynamic_cast<const CFunc3*>( fun ) ) {
            if( ctx.want_error && fu->d3 && err_input ) {
                if ( is_scalar ) {
                    (*(fu->d3))( ret.r, ret.dr, a[0].r, a[0].dr,
                                 a[1].r, a[1].dr, a[2].r, a[2].dr );
                } else {
                    for ( int i=0; i<n; ++i )
                        (*(fu->d3))(
                            ret.v[i], ret.dv[i],
                            a[0].to_r(i), a[0].to_dr(i),
                            a[1].to_r(i), a[1].to_dr(i),
                            a[2].to_r(i), a[2].to_dr(i) );
                }
            } else {
                if ( is_scalar ) 
                    ret.set_r( (*(fu->f3))( a[0].r, a[1].r, a[2].r ) );
                else {
                    for ( int i=0; i<n; ++i )
                        ret.v[i] = (*(fu->f3))(
                            a[0].to_r(i), a[1].to_r(i), a[2].to_r(i) );
                }
            }
        } else
            throw string( "BUG: unexpected subclass of CFunc" );
        if( DEBUG ) {
            cout << "tree_op:\n";
            for ( int iarg=0; iarg<narg; ++iarg )
                cout << "  arg: " << a[iarg].result_info() << "\n";
            cout << "  res: " << ret.result_info() << "\n";
        }
    } catch( string& ex ) {
        throw ex + " in NodeFun " + tree_info() +
            " {context=" + ctx.context_info() + "}";
    }
}

bool CNodeFun::has_dummy() const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        if( arg[iarg]->has_dummy() )
            return true;
    return false;
}

bool CNodeFun::has_conv() const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        if( arg[iarg]->has_conv() )
            return true;
    return false;
}

void CNodeFun::kdep_exec( bool *kd ) const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        arg[iarg]->kdep_exec( kd );
}

void CNodeFun::npar_exec( int *np ) const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        arg[iarg]->npar_exec( np );
}

void CNodeFun::set_coord( CCoord& ret, int k ) const {
    CCoord r[maxarg];
    for ( int iarg=0; iarg<narg; ++iarg )
        arg[iarg]->set_coord( r[iarg], k );
    if      ( const CFunc1* fu = dynamic_cast<const CFunc1*>( fun ) )
        ret = fu->coord( r+0);
    else if ( const CFunc2* fu = dynamic_cast<const CFunc2*>( fun ) )
        ret = fu->coord( r+0, r+1 );
    else if ( const CFunc3* fu = dynamic_cast<const CFunc3*>( fun ) )
        ret = fu->coord( r+0, r+1, r+2 );
}

//! Returns string representation of this node.

string CNodeFun::tree_info() const
{
    string ret = fun->txt + "(";
    for ( int iarg=0; iarg<narg; ++iarg )
        ret += arg[iarg]->tree_info() + (iarg<narg-1 ? "," : ")" );
    return ret;
}


//**************************************************************************************************
//* CNodeGeni: generalized integral node                          
//**************************************************************************************************


CNodeGeni::CNodeGeni( const class CGeni *_geni, PRef _ref )
    : narg(_geni->narg), geni(_geni)
{
    arg.resize(narg);
    for ( int iarg=0; iarg<narg; ++iarg )
        arg[iarg] = geni->def_arg[iarg];
}

CNodeGeni::CNodeGeni( const class CGeni *_geni, PNode _a0 )
    : narg(1), geni(_geni)
{
    if ( geni->narg!=1 )
        throw "syntax error: generalized integral '" + geni->name + "' needs " +
            S(geni->narg) + " arguments, not 1";
    arg.resize(narg);
    arg[0] = _a0;
}

CNodeGeni::CNodeGeni( const class CGeni *_geni, PNode _a0, PNode _a1 )
    : narg(2), geni(_geni)
{
    if ( geni->narg!=2 )
        throw "syntax error: generalized integral '" + geni->name + "' needs " +
            S(geni->narg) + " arguments, not 2";
    arg.resize(narg);
    arg[0] = _a0;
    arg[1] = _a1;
}

//! A functional of 1,2,.. arguments (including operators like +,-,.. ).

void CNodeGeni::tree_val( CResult& ret, const CContext& ctx ) const
{
    try {
        if( ctx.dim!=CContext::_1 )
            throw "Unexpected vectorial context " + ctx.context_info();
        CContext myctx(ctx);
        myctx.dim = CContext::_VI;
        myctx.nv = -1;
        // evaluate arguments,
        vector<PResult> a(narg);
        bool has_error = false;
        for ( int iarg=0; iarg<narg; ++iarg ) {
            a[iarg] = PResult( new CResult() );
            arg[iarg]->tree_val( *(a[iarg]), myctx );
            if ( a[iarg]->has_err() )
                has_error = true;
        }
        // determine size of vectorial arguments
        int n=0;
        for ( int iarg=0; iarg<narg; ++iarg ) {
            if ( !a[iarg]->vectorial )
                throw "argument " + S(iarg) + " of generalized integral " +
                    tree_info() + " is scalar: " + a[iarg]->result_info();
            if ( n==0 )
                n = a[iarg]->v.size();
            else if ( a[iarg]->v.size() != n )
                throw string("vector arguments have different size");
        }
        // now evaluate the function
        ret.vectorial = false;
        geni->eval( &(ret.r), (ctx.want_error && has_error) ? &(ret.dr) : NULL,
              n, a );
    } catch( string& ex ) {
        throw ex + " in NodeGeni " + tree_info() +
            " {context=" + ctx.context_info() + "}";
    }
}

bool CNodeGeni::has_dummy() const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        if( arg[iarg]->has_dummy() )
            return true;
    return false;
}

bool CNodeGeni::has_conv() const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        if( arg[iarg]->has_conv() )
            return true;
    return false;
}

void CNodeGeni::kdep_exec( bool *kd ) const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        arg[iarg]->kdep_exec( kd );
}

void CNodeGeni::npar_exec( int *np ) const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        arg[iarg]->npar_exec( np );
}

void CNodeGeni::set_coord( CCoord& ret, int k ) const {
//*     
//*     CCoord r[maxarg];
//*     for ( int iarg=0; iarg<narg; ++iarg )
//*         arg[iarg]->set_coord( r[iarg], k );
//*     if      ( const CFunc1* fu = dynamic_cast<const CFunc1*>( fun ) )
//*         ret = fu->coord( r+0);
//*     else if ( const CFunc2* fu = dynamic_cast<const CFunc2*>( fun ) )
//*         ret = fu->coord( r+0, r+1 );
//*     else if ( const CFunc3* fu = dynamic_cast<const CFunc3*>( fun ) )
//*         ret = fu->coord( r+0, r+1, r+2 );
}


//! Returns string representation of this node.

string CNodeGeni::tree_info() const
{
    string ret = geni->name+"(";
    for ( int iarg=0; iarg<narg; ++iarg )
        ret += arg[iarg]->tree_info() + (iarg<narg-1 ? "," : ")" );
    return ret;
}


//**************************************************************************************************
//* CNodeCvin: curve functional node                          
//**************************************************************************************************


CNodeCvin::CNodeCvin(
    const class CCvin *_cvin, PRef _ref, PNode _a0, PNode _a1 )
    : narg(2), cvin(_cvin), ref(_ref)
{
    if ( cvin->narg!=2 )
        throw "syntax error: generalized integral '" + cvin->txt + "' needs " +
            S(cvin->narg) + " arguments, not 2";
    arg.resize(narg);
    arg[0] = _a0;
    arg[1] = _a1;
}

void CNodeCvin::tree_val( CResult& ret, const CContext& ctx ) const
{
    try {
        if( ctx.dim!=CContext::_1 )
            throw "Unexpected vectorial context " + ctx.context_info();
        CContext myctx(ctx);
        myctx.dim = CContext::_1;
        // evaluate arguments,
        vector<double> a(narg);
        bool has_error = false;
        for ( int iarg=0; iarg<narg; ++iarg ) {
            CResult resa;
            arg[iarg]->tree_val( resa, myctx );
            if ( resa.has_err() )
                has_error = true;
            if ( resa.vectorial )
                throw "argument " + S(iarg) + " of curve functional " +
                    tree_info() + " is vectorial: " + resa.result_info();
            a[iarg] = resa.r;
        }
        int k = ref->get_k( ctx );
        POlo f = NOlm::mem_get(k);
        POlc fc = P2C( f );
        if ( !fc )
            throw string("This is not a curve file => cannot integrate");
        int j = ref->get_j( ctx, f->nJ() );
        // now evaluate the function
        ret.vectorial = false;
        cvin->eval( &(ret.r), (ctx.want_error && has_error) ? &(ret.dr) : NULL,
                    fc, k, j, a );
    } catch( string& ex ) {
        throw ex + " in NodeCvin " + tree_info() +
            " {context=" + ctx.context_info() + "}";
    }
}

bool CNodeCvin::has_dummy() const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        if( arg[iarg]->has_dummy() )
            return true;
    return false;
}

bool CNodeCvin::has_conv() const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        if( arg[iarg]->has_conv() )
            return true;
    return false;
}

void CNodeCvin::kdep_exec( bool *kd ) const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        arg[iarg]->kdep_exec( kd );
}

void CNodeCvin::npar_exec( int *np ) const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        arg[iarg]->npar_exec( np );
}

void CNodeCvin::set_coord( CCoord& ret, int k ) const {
//*     
//*     CCoord r[maxarg];
//*     for ( int iarg=0; iarg<narg; ++iarg )
//*         arg[iarg]->set_coord( r[iarg], k );
//*     if      ( const CFunc1* fu = dynamic_cast<const CFunc1*>( fun ) )
//*         ret = fu->coord( r+0);
//*     else if ( const CFunc2* fu = dynamic_cast<const CFunc2*>( fun ) )
//*         ret = fu->coord( r+0, r+1 );
//*     else if ( const CFunc3* fu = dynamic_cast<const CFunc3*>( fun ) )
//*         ret = fu->coord( r+0, r+1, r+2 );
}

//! Returns string representation of this node.

string CNodeCvin::tree_info() const
{
    string ret = cvin->txt+"(..)";
    for ( int iarg=0; iarg<narg; ++iarg )
        ret += arg[iarg]->tree_info() + (iarg<narg-1 ? "," : ")" );
    return ret;
}


//**************************************************************************************************
//* CNodeVal: end node containing a numeric value             
//**************************************************************************************************

CNodeVal::CNodeVal( double _val )
    : val( _val ) {}

void CNodeVal::set_coord( CCoord& ret, int k ) const {
        ret = CCoord(S(val), ""); }
string CNodeVal::tree_info() const { return "val[" + S(val) + "]"; }


//**************************************************************************************************
//* CNodeIva: indexed variable node                           
//**************************************************************************************************

void CNodeIva::npar_exec( int *np ) const
{
    if ( var->typ == CVar::_CP )
        *np = max( *np, var->num+1 );
}

//! Evaluate reference in given context. Result returned in ret.

void CNodeIva::tree_val( CResult& ret, const CContext& ctx ) const
{
    try {
        if ( (var->typ == CVar::_I || var->typ == CVar::_J ||
              var->typ == CVar::_K) && ( ref->ti || ref->tj || ref->tk ) )
            throw string("invalid cross reference: index cannot be indexed");

        // Get k from context:
        int k = ref->get_k( ctx );
        
        // Return k-dependent reference (unless already done above):
        if ( var->typ == CVar::_K ) {
            ret.set_r( k );
            return;
        }

        // Proceed with file reference f[k]:
        POlo f = NOlm::mem_get(k);

        if        ( var->typ == CVar::_NJ ) {
            ret.set_r( f->nJ() );
            return;
        } else if ( var->typ == CVar::_R ) {
            if ( var->num>= f->RPar.size() )
                throw string("invalid reference ") + var->var_info();
            ret.set_r( f->RPar[var->num].val );
            return;
        }

        // Return k-j-dependent reference (unless done above):
        if        ( var->typ == CVar::_J ) {
            if ( ref->ti )
                throw string("J does not require 3 indices");
            ret.set_r( ctx.j );
            return;
        } else if ( var->typ == CVar::_Z ) {
            if ( ref->ti )
                throw string("Z does not require 3 indices");
            if ( var->num>= f->nZ() )
                throw string("invalid reference ") + var->var_info();
            int j = ref->get_j( ctx, f->nJ() );
            ret.set_r( f->V[j]->z[var->num] );
            return;
        }

        // References for data file
        POld fd = P2D( f );
        if ( fd ) {
            if        ( var->pointwise() ) {
                if        ( ctx.dim==CContext::_1 ) {
                    int j = ref->get_j( ctx, f->nJ() );
                    PSpec sj = fd->VS(j);
                    int i;
                    if ( ref->ti ) {
                        ref->ti->tree_uival( &i, ctx );
                    } else {
                        i = ctx.i;
                        if ( i==(int)-1 )
                            throw string("missing index i");
                    }
                    if( i>= sj->size() )
                        throw "i=" + S(i) + " while there are only ni="
                            + S(sj->size()) + " points";
                    if       ( var->typ == CVar::_I ) {
                        ret.set_r( i );
                    } else if ( var->typ == CVar::_X ) {
                        ret.set_r( sj->x[i] );
                    } else if ( var->typ == CVar::_Y ) {
                        ret.set_r( sj->y[i] );
                        if ( ctx.want_error && sj->dy.size() )
                            ret.dr = sj->dy[i];
                    } else if ( var->typ == CVar::_DY ) {
                        ret.set_r( sj->dy[i] );
                    }

                } else if ( ctx.dim==CContext::_VI ) {
                    if ( ref->ti ) {
                        CContext myctx = ctx;
                        myctx.dim = CContext::_1;
                        ret.preset_v( ctx.nv, false );
                        if ( var->typ != CVar::_Y && ctx.want_error )
                            ret.dv.resize( ctx.nv, 0. );
                        for( int ii=0; ii<ctx.nv; ++ii ){
                            myctx.i = ii;
                            int j = ref->get_j( myctx, f->nJ() );
                            PSpec sj = fd->VS(j);
                            int i;
                            ref->ti->tree_uival( &i, myctx );
                            if( i >= sj->size() )
                                throw "i=" + S(i) + " while there are only ni="
                                    + S(sj->size()) + " points";
                            if        ( var->typ == CVar::_I ) {
                                ret.v[ii] = i;
                            } else if ( var->typ == CVar::_X ) {
                                ret.v[ii] = sj->x[i];
                            } else if ( var->typ == CVar::_Y ) {
                                ret.v[ii] = sj->y[i];
                                if( ret.dv.size() ){
                                    if ( sj->dy.size() >= ret.dv.size() ) 
                                        ret.dv[ii] = sj->dy[i];
                                    else
                                        ret.dv.clear();
                                }
                            } else if ( var->typ == CVar::_DY ) {
                                if( !sj->dy.size() )
                                    throw string("dy not set");
                                ret.v[ii] = sj->dy[i];
                            }
                        }
                        
                    } else {
                        int j = ref->get_j( ctx, f->nJ() );
                        PSpec sj = fd->VS(j);
                        int nv = ctx.nv;
                        if ( nv==-1 ) {
                            nv = sj->size();
                        } else if ( nv != sj->size() )
                            throw string("ref_val ") + var->var_info() +
                                ": requested vec(" + S(ctx.nv) +
                                "), found vec(" + S( sj->size() );
                        if        ( var->typ == CVar::_I ) {
                            ret.preset_v( nv, false );
                            for( int i=0; i<nv; ++i )
                                ret.v[i] = i;
                        } else if ( var->typ == CVar::_X ) {
                            ret.set_v( sj->x );
                        } else if ( var->typ == CVar::_Y ) {
                            ret.set_v( sj->y );
                            if ( ctx.want_error )
                                ret.dv = sj->dy;
                        } else if ( var->typ == CVar::_DY ) {
                            ret.set_v( sj->dy );
                        }
                    }
                } else
                    throw string("BUG: invalid context::dim");

            } else if ( var->typ == CVar::_NI ) {
                int j = ref->get_j( ctx, f->nJ() );
                PSpec sj = fd->VS(j);
                ret.set_r( sj->size() );

            } else
                throw string("invalid ref(") + var->var_info() +
                    ") in data file";

            return;
        }

        // References for curve file:
        POlc fc = P2C( f );
        if ( fc ) {
            if ( ref->ti )
                throw string("unexpected i dependence of curve property");
            int j = ref->get_j( ctx, f->nJ() );
            PCurve cj = fc->VC(j);
            if       ( var->typ == CVar::_CP ) {
                if ( var->num>= fc->nP )
                    throw string("invalid p ref(") + var->var_info() +
                        ") in curve file";
                ret.set_r( cj->P[var->num] );

            } else if ( var->typ == CVar::_COUTCOME ) {
                ret.set_r( cj->fitOutcome );
            } else if ( var->typ == CVar::_CCHI2 ) {
                ret.set_r( cj->fitChi2 );
            } else if ( var->typ == CVar::_CR2 ) {
                ret.set_r( cj->fitR2 );

            } else 
                throw string("invalid ref(") + var->var_info() +
                    ") in curve file";

            return;
        }
    } catch( string& ex ) {
        throw ex + " in NodeIva " + tree_info() +
            " {context=" + ctx.context_info() + "}";
    }
}

//! Description of reference (for use in ??)

void CNodeIva::set_coord( CCoord& ret, int k_in ) const
{
    if        ( var->typ == CVar::_K ) {
        ret = CCoord("k","");
        return;
    } else if ( var->typ == CVar::_J ) {
        ret = CCoord("j","");
        return;
    } else if ( var->typ == CVar::_I ) {
        ret = CCoord("i","");
        return;
    }

    int k;
    if (ref->tk) {
        // get k from reference :
        CContext ctx( k_in );
        ref->tk->tree_uival(&k, ctx);
    } else {
        k = k_in;
    }
    POlo f = NOlm::mem_get(k);

    if  ( var->typ == CVar::_Z ) {
        if (var->num>=f->ZCo.size())
            throw "set_coord: z" + S(var->num) + " undefined";
        ret = f->ZCo[var->num];
    } else if  ( var->typ == CVar::_R ) {
        if (var->num>=f->RPar.size())
            throw "set_coord: r" + S(var->num) + " undefined";
        ret = f->RPar[var->num].Co;
    } else if ( var->typ == CVar::_NJ ) {
        ret = CCoord("#spectra", "");
    } else {
        POld fd = P2D( f );
        POlc fc = P2C( f );
        if         ( fd ) {
            if       ( var->typ == CVar::_X )
                ret = fd->xco;
            else if  ( var->typ == CVar::_Y )
                ret = fd->yco;
            else if  ( var->typ == CVar::_DY )
                ret = CCoord( "d"+fd->yco.name, fd->yco.unit );
            else if  ( var->typ == CVar::_NI )
                ret = CCoord("#points", "");
            else
                throw "data file has no " + var->var_info();
        } else if ( fc ) {
            if     ( var->typ == CVar::_CP ) {
                if( var->num>=fc->nP )
                    throw "invalid reference p" + S(var->num);
                ret = fc->PCo[var->num];
            }
            else if ( var->typ == CVar::_COUTCOME )
                ret = CCoord("fit_outcome", "");
            else if ( var->typ == CVar::_CCHI2 )
                ret = CCoord("chi2", "");
            else if ( var->typ == CVar::_CR2 )
                ret = CCoord("R2", "");
            else if ( var->typ == CVar::_C )
                ret = CCoord(fc->expr,"");
            else
                throw "curve file has no " + var->var_info();
        } else
            throw "BUG: ref->set_coord unexpected else";
    }
}

//! Returns string representation of this node.

string CNodeIva::tree_info() const
{
    return "ref[" + var->var_info() + "]";
}


//**************************************************************************************************
//* CNodeRgr: register reference node                         
//**************************************************************************************************

CNodeRgr::CNodeRgr( PRgr& _rgr )
    : rgr( _rgr ) {}

void CNodeRgr::tree_val( CResult& ret, const CContext& ctx ) const {
    ret.set_r( NReg::lastresult ); }
void CNodeRgr::set_coord( CCoord& ret, int k ) const {
    ret = CCoord( "$", "" ); } // primitiver gehts nicht


//**************************************************************************************************
//* CNodeDummy: function argument (t) node                    
//**************************************************************************************************

// .. nothing to do ..


//**************************************************************************************************
//* CNodeCev: curve evaluation node                           
//**************************************************************************************************

CNodeCev::CNodeCev( PRef& _ref, PNode _arg )
    : ref( _ref ), arg( _arg )
{
    if( ref->ti )
        throw "curve reference must not depend on index i";
}

//! A curve evaluation: fc[k,j](arg).

void CNodeCev::tree_val( CResult& ret, const CContext& ctx ) const
{
    try {
        // get k from context:
        int k = ref->get_k( ctx );

        // k points to a curve:
        POlc fc = NOlm::mem_get_C( k );

        // get j from context:
        int j = ref->get_j( ctx, fc->nJ() );

        // evaluate and pack function argument:
        CResult farg;
        arg->tree_val( farg, ctx );

        // evaluate the curve:
        if ( farg.vectorial ) {
            ret.preset_v( farg.v.size(), ctx.want_error );
            fc->curve_val_vec( &(ret.v), ctx.want_error? &(ret.dv) : NULL,
                               farg.v, k, j );
        } else {
            ret.vectorial = false;
            fc->curve_val_sca( &(ret.r), ctx.want_error? &(ret.dr) : NULL,
                               farg.r, k, j );
        }
    } catch( string& ex ) {
        throw ex + " in NodeCev " + tree_info() +
            " {context=" + ctx.context_info() + "}";
    }
}

void CNodeCev::set_coord( CCoord& ret, int k ) const {
    ret = CCoord( "c", "" ); } // primitiver gehts nicht


//**************************************************************************************************
//* CNodeMixin: Base class for mixing in data from a "convolution" file
//**************************************************************************************************

void CNodeMixin::tree_val( CResult& ret, const CContext& ctx ) const
{
    // Shift:
    CResult res_shift;
    shift->tree_val( res_shift, ctx );
    if ( res_shift.vectorial )
        throw "shift is not a scalar";
    double theshift = res_shift.r;

    // Return values are requested for n values of _DUMMY ( = t ):
    if ( !(ctx.vt) )
        throw "BUG: convolution tree_val with *vt=0";
    int n = ctx.vt->size();
    ret.preset_v( n, ctx.want_error );

    // Resolution is given by context 'ctx':
    int kconv, jconv;
    PSpec sv;
    NCurveFile::getConv( &sv, &kconv, &jconv, ctx.k, ctx.j );

    if ( kconv==-1 ) {
        copy_theory( ret, ctx, theshift );
    } else {
        double conv_norm; // normalization constant (integral of resolution)
        double conv_step; // step in x/t, if equidistant
        string info = "resolution " + S(kconv) + ", spectrum " + S(jconv);
        conv_norm = sv->norm( info );
        conv_step = sv->has_step( ctx.vt_step ) ? ctx.vt_step : 0;
        if( !conv_step && !allow_slow_conv )
            throw "grid is not equidistant; slow convolution is not allowed";
        convolve( ret, ctx, theshift, sv, conv_norm, conv_step );
    }
}


//**************************************************************************************************
//* CNodeConvBase: Base class for convoluting a function with a spectrum.
//**************************************************************************************************

void CNodeConvBase::kdep_exec( bool *kd ) const
{
    theory->kdep_exec( kd );
    shift->kdep_exec( kd );
}

void CNodeConvBase::npar_exec( int *np ) const
{
    theory->npar_exec( np );
    shift->npar_exec( np );
}


//**************************************************************************************************
//* CNodeConv: convoluted function node                       
//**************************************************************************************************

void CNodeConv::copy_theory( CResult& ret, const CContext& ctx,
                             double theshift ) const
{
    CResult res_theory;
    theory->tree_val( res_theory, ctx );
    for ( int i=0; i<ctx.vt->size(); ++i )
        ret.v[i] = res_theory.to_r(i);
}

void CNodeConv::convolve( CResult& ret, const CContext& ctx,
                          double theshift, const PSpec& sv,
                          double conv_norm, double conv_step ) const
{
    int n = ctx.vt->size();
    int nv = sv->size();
    CResult res_theory;
    CContext cv_ctx(ctx);
    vector<double> tt; // theory grid
    cv_ctx.vt = &tt;
    bool with_error = ctx.want_error && sv->dy.size();
    if( with_error ) {
        if( ret.dv.size()!=n )
            throw "BUG: convolve: bad #dv";
        if( sv->dy.size()!=n )
            throw "BUG: convolve: bad #dy";
    }
    // Convolute theory with resolution.
    if( conv_step ){ // accelerated version with equidistant grids
        // precompute theory on unified grid
        int ntt = n + nv - 1;
        tt.resize(ntt);
        double tt0 = (*(ctx.vt))[0] - sv->x[0] - theshift - (nv-1)*conv_step;
        for ( int ii=0; ii<ntt; ++ii ) // ii=(nv-1)+i-iv
            tt[ii] = tt0 + ii*conv_step;
        theory->tree_val( res_theory, cv_ctx );
        // now convolve
        for ( int i=0; i<n; ++i ){
            ret.v[i] = 0;
            if( with_error )
                ret.dv[i] = 0;
            for ( int iv=0; iv<nv; ++iv ) {
                ret.v[i] += res_theory.to_r(nv-1+i-iv) * sv->y[iv];
                if( with_error )
                    ret.dv[i] += res_theory.to_r(nv-1+i-iv) * SQR(sv->dy[iv]);
            }
            ret.v[i] *= conv_step / conv_norm;
            if( with_error )
                ret.dv[i] = sqrt(ret.dv[i]) * conv_step / conv_norm;
        }
    } else { // simplest version: non-equidistant x_out and x_res.
        ret.v.assign( n, 0. );
        if( with_error )
            ret.dv.assign( n, 0. );
        tt.resize(n);
        for ( int iv=0; iv<nv; ++iv ) {
            double dx; // variable step
            if      ( iv==0 )
                dx = sv->x[1] - sv->x[0];
            else if ( iv==nv-1 )
                dx = sv->x[nv-1] - sv->x[nv-2];
            else
                dx = ( sv->x[iv+1] - sv->x[iv-1] ) / 2;
            for ( int i=0; i<n; ++i )
                tt[i] = (*(ctx.vt))[i] - sv->x[iv] - theshift;
            theory->tree_val( res_theory, cv_ctx );
            for ( int i=0; i<n; ++i ) {
                ret.v[i] += res_theory.to_r(i) * sv->y[iv] * dx;
                if( with_error )
                    ret.dv[i] += res_theory.to_r(i) * SQR(sv->dy[iv]);
            }
        }
        for ( int i=0; i<n; ++i ) {
            ret.v[i] /= conv_norm;
            if( with_error )
                ret.dv[i] = sqrt(ret.dv[i])/conv_norm;
        }
    }
}


//**************************************************************************************************
//* CNodePConv: convolution with primitive                    
//**************************************************************************************************

void CNodePConv::copy_theory( CResult& ret, const CContext& ctx,
                              double theshift ) const
{
    throw "non-convolution not implemented for pconv";
    // to implement this efficiently, I would like to have CResult::operator+
    // for which we should go to shared pointers ...
}

void CNodePConv::convolve( CResult& ret, const CContext& ctx,
                          double theshift, const PSpec& sv,
                          double conv_norm, double conv_step ) const
{
    int n = ctx.vt->size();
    int nv = sv->size();
    CContext cv_ctx(ctx);
    vector<double> tt;
    cv_ctx.vt = &tt;
    bool with_error = ctx.want_error && sv->dy.size();
    if( with_error ) {
        if( ret.dv.size()!=n )
            throw "BUG: convolve: bad #dv";
        if( sv->dy.size()!=n )
            throw "BUG: convolve: bad #dy";
    }
    // Convolute theory with resolution.
    if( conv_step ){ // accelerated version with equidistant grids
        // precompute primitive on unified grid
        int ntt = n + nv;
        tt.resize(ntt);
        double tt0 = (*(ctx.vt))[0] - sv->x[0] - theshift - (nv-0.5)*conv_step;
        for ( int ii=0; ii<ntt; ++ii ) // ii=(nv-1)+i-iv
            tt[ii] = tt0 + ii*conv_step;
        CResult res_theory;
        theory->tree_val( res_theory, cv_ctx );
        // now convolve resolution with differential of primitive
        for ( int i=0; i<n; ++i ){
            ret.v[i] = 0;
            if( with_error )
                ret.dv[i] = 0;
            for ( int iv=0; iv<nv; ++iv ) {
                double igral =
                    res_theory.to_r(nv  +i-iv)-
                    res_theory.to_r(nv-1+i-iv);
                ret.v[i] += igral * sv->y[iv];
                if( with_error )
                    ret.dv[i] += igral * SQR(sv->dy[iv]);
            }
            ret.v[i] /= conv_norm;
            if( with_error )
                ret.dv[i] = sqrt(ret.dv[i]) / conv_norm;
        }
    } else { // simplest version: non-equidistant x_out and x_res.
        ret.v.assign( n, 0. );
        if( with_error )
            ret.dv.assign( n, 0. );
        tt.resize(n);
        CResult res_theory_low, res_theory_hig;
        double dx;
        for ( int iv=0; iv<nv; ++iv ) {
            if      ( iv==0 )
                dx = sv->x[1] - sv->x[0];
            else if ( iv==nv-1 )
                dx = sv->x[nv-1] - sv->x[nv-2];
            else
                dx = ( sv->x[iv+1] - sv->x[iv-1] ) / 2;
            for ( int i=0; i<n; ++i ) {
                tt[i] = (*(ctx.vt))[i] - sv->x[iv] - theshift - dx/2;
                theory->tree_val( res_theory_low, cv_ctx );
            }
            for ( int i=0; i<n; ++i ) {
                tt[i] = (*(ctx.vt))[i] - sv->x[iv] - theshift + dx/2;
                theory->tree_val( res_theory_hig, cv_ctx );
            }
            for ( int i=0; i<n; ++i ) {
                double igral = res_theory_hig.to_r(i)-res_theory_low.to_r(i);
                ret.v[i] += igral * sv->y[iv] * dx;
                if ( with_error )
                    ret.dv[i] += igral * SQR(sv->dy[iv]) * dx;
            }
        }
        for ( int i=0; i<n; ++i ) {
            ret.v[i] /= conv_norm;
            if ( with_error )
                ret.dv[i] = sqrt(ret.dv[i]) / conv_norm;
        }
    }
}


//**************************************************************************************************
//* CNodeDirac: node that copies the resolution function      
//**************************************************************************************************

void CNodeDirac::kdep_exec( bool *kd ) const
{
    shift->kdep_exec( kd );
}

void CNodeDirac::npar_exec( int *np ) const
{
    shift->npar_exec( np );
}

//! Convolution or Dirac function: conv(f, shift), dirac(shift).

void CNodeDirac::convolve( CResult& ret, const CContext& ctx,
                           double theshift, const PSpec& sv,
                           double conv_norm, double conv_step ) const
{
    int n = ctx.vt->size();
    bool with_error = ctx.want_error && sv->dy.size();
    if( with_error ) {
        if( ret.dv.size()!=n )
            throw "BUG: convolve: bad #dv";
        if( sv->dy.size()!=n )
            throw "BUG: convolve: bad #dy";
    }
    // Delta function copies resolution.
    vector<double> tt(n), yy(n), dd(n);
    // shift in t
    for ( int i=0; i<n; ++i )
        tt[i] = (*(ctx.vt))[i]-theshift;
    // interpolation
    try {
        sv->intpol( tt, &yy, with_error ? &dd : NULL );
    } catch( string& s ) {
        throw "error while evaluating delta function with shift " +
            S(theshift) + ": " + s;
    }
    // normalization
    for ( int i=0; i<n; ++i ) {
        ret.v[i] = yy[i] / conv_norm;
        if( with_error )
            ret.dv[i] = dd[i] / conv_norm;
    }
}

void CNodeDirac::copy_theory( CResult& ret, const CContext& ctx,
                              double theshift ) const
{
    // Naive representation of delta function,
    // for use when convolution is switched off.
    for ( int i=0; i<ctx.vt->size(); ++i )
        ret.v[i] = fabs((*(ctx.vt))[i]-theshift)<1e-2 ? 1e2 : 0;
}

void CNodeDirac::set_coord( CCoord& ret, int k ) const
{
    ret = CCoord("resol", "");
}
