//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis                          
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-                                 
//*  http://apps.jcns.fz-juelich.de/frida                                    
//**************************************************************************************************

//! \file  olf.cpp
//! \brief COlo, COld, COld online files

#include <fstream>
#include <iostream>
#include <sstream>
#include <exception>
#include <fcntl.h>
#include <sys/stat.h>
#include <boost/format.hpp>

#include <trivia/file_ops.hpp>
#include <readplus/ask.hpp>

#include "defs.hpp"
#include "olf.hpp"
#include "slice.hpp"
#include "var.hpp"
#include "expr.hpp"
#include "xax_lex.hpp"

using boost::format;

//**************************************************************************************************
//*  class COlo                                                              
//**************************************************************************************************

//! Remove one z.

void COlo::remove_z( int iz )
{
    ZCo.erase( ZCo.begin()+iz );
    for( int j=0; j<nJ(); ++j )
        V[j]->z.erase( V[j]->z.begin()+iz );
}

//! Line containing z values and curve parameters.

string COlo::infoLine( int j ) const
{
    const int LEN=15;
    char wrd[LEN];
    snprintf( wrd, LEN, "%-2i%c ", j, V[j]->frozen ? '/' : ' ' );
    string out = wrd;
    out += V[j]->infoLine();
    return out;
}

//! Return value of iz-th z-variable for slice j.

double COlo::z( int j, int iz ) const
{
    return V[j]->z[iz];
}

//**************************************************************************************************
//*  trivially duplicated COld/COlc functions                                
//**************************************************************************************************

//! Duplicate spectrum j, and return new shared pointer.

PSlice COld::new_slice( int j ) const
{
    return PSlice( new CSpec( *(VS(j)) ) );
}

//! Duplicate curve j, and return new shared pointer.

PSlice COlc::new_slice( int j ) const
{
    return PSlice( new CCurve( *(VC(j)) ) );
}

//! New shared pointer to copied online file (but without copying main vector).

POlo COld::new_olo( bool modified ) const
{
    POlo ret( new COld( *this ) );
    ret->V.clear();
    ret->as_on_disk = as_on_disk && !modified;
    return ret;
}

POlo COlc::new_olo( bool modified ) const
{
    POlo ret( new COlc( *this ) );
    ret->V.clear();
    ret->as_on_disk = as_on_disk && !modified;
    return ret;
}

POld COld::new_old( bool modified ) const
{
    return P2D( new_olo( modified ) );
}

POlc COlc::new_olc( bool modified ) const
{
    return P2C( new_olo( modified ) );
}

//! Set main vector by copying from fin.

void COld::copy_mainvec( POlo fin )
{
    POld fd = boost::dynamic_pointer_cast<class COld>( fin );
    V.resize( fin->nJ() );
    for( int j=0; j<fin->nJ(); ++j )
        V[j] = PSlice( new CSpec( *(fd->VS(j)) ) );
}

//! Set main vector by copying from fin.

void COlc::copy_mainvec( POlo fin )
{
    POlc fc = boost::dynamic_pointer_cast<class COlc>( fin );
    V.resize( fin->nJ() );
    for( int j=0; j<fin->nJ(); ++j )
        V[j] = PSlice( new CCurve( *(fc->VC(j)) ) );
}


//! Return pointer to selected main vector entry.

PSpec COld::VS( int j ) const
{
    if( j>=V.size() )
        throw "VS call with j=" + S(j) + " though V.size=" + S(V.size());
    PSpec s = boost::dynamic_pointer_cast<class CSpec>( V[j] );
    if ( !s )
        throw "invalid VS call";
    return s;
}

PCurve COlc::VC( int j ) const
{
    if( j>=V.size() )
        throw "VC call with j=" + S(j) + " though V.size=" + S(V.size());
    PCurve c = boost::dynamic_pointer_cast<class CCurve>( V[j] );
    if ( !c )
        throw "invalid VC call";
    return c;
}


//! Return one coordinate.

CCoord COld::coord( CVar* var ) const
{
    if        ( var->typ==CVar::_X ) {
        return xco;
    } else if ( var->typ==CVar::_Y ) {
        return yco;
    } else if ( var->typ==CVar::_Z ) {
        if( var->num >= nZ() )
            throw "coordinate " + var->var_info() + " not defined";
        return ZCo[var->num];
    } else if ( var->typ==CVar::_R ) {
        if( var->num >= RPar.size() )
            throw "coordinate " + var->var_info() + " not defined";
        return RPar[var->num].Co;
    } else {
        throw "variable " + var->var_info() + " not defined for curve file";
    }
}

//! Return one coordinate.

CCoord COlc::coord( CVar* var ) const
{
    if        ( var->typ==CVar::_X ) {
        return xco;
    } else if ( var->typ==CVar::_Y ) {
        return yco;
    } else if ( var->typ==CVar::_Z ) {
        if( var->num >= nZ() )
            throw "coordinate " + var->var_info() + " not defined";
        return ZCo[var->num];
    } else if ( var->typ==CVar::_CP ) {
        if( var->num >= nP )
            throw "coordinate " + var->var_info() + " not defined";
        return PCo[var->num];
    } else if ( var->typ==CVar::_R ) {
        if( var->num >= RPar.size() )
            throw "coordinate " + var->var_info() + " not defined";
        return RPar[var->num].Co;
    } else {
        throw "variable " + var->var_info() + " not defined for curve file";
    }
}

//! Change one coordinate.

void COld::setCoord( CVar* var, CCoord& co )
{
    if        ( var->typ==CVar::_X ) {
        xco = co;
    } else if ( var->typ==CVar::_Y ) {
        yco = co;
    } else if ( var->typ==CVar::_Z ) {
        if( var->num >= nZ() )
            throw "coordinate " + var->var_info() + " not defined";
        ZCo[var->num] = co;
    } else if ( var->typ==CVar::_R ) {
        if( var->num >= RPar.size() )
            throw "coordinate " + var->var_info() + " not defined";
        RPar[var->num].Co = co;
    } else {
        throw "variable " + var->var_info() + " not defined for curve file";
    }
}

//! Change one coordinate.

void COlc::setCoord( CVar* var, CCoord& co )
{
    if        ( var->typ==CVar::_X ) {
        xco = co;
    } else if ( var->typ==CVar::_Y ) {
        yco = co;
    } else if ( var->typ==CVar::_Z ) {
        if( var->num >= nZ() )
            throw "coordinate " + var->var_info() + " not defined";
        ZCo[var->num] = co;
    } else if ( var->typ==CVar::_CP ) {
        if( var->num >= nP )
            throw "coordinate " + var->var_info() + " not defined";
        PCo[var->num] = co;
    } else if ( var->typ==CVar::_R ) {
        if( var->num >= RPar.size() )
            throw "coordinate " + var->var_info() + " not defined";
        RPar[var->num] = co;
    } else {
        throw "variable " + var->var_info() + " not defined for curve file";
    }
}

//**************************************************************************************************
//*  class COld                                                              
//**************************************************************************************************

//! Create data file based on curve file.

COld::COld( class COlc const* c )
{
    name = c->name;
    lDoc = c->lDoc;
    xco = c->xco;
    yco = c->yco;
    ZCo = c->ZCo;
    RPar = c->RPar;
}

//! Return number of points in spectrum j.

int COld::nPts( int j ) const {
    if ( !nJ() || j>=nJ() )
        throw "BUG: nPts called with invalid j="+S(j);
    return VS(j)->size();
}


//! Return number of points if it is the same for all spectra. Else return 0.

int COld::nPts() const {
    if ( !nJ() )
        throw "BUG: nPts() called while nSpec=0";
    int np = VS(0)->size();
    for ( int j=0; j<nJ(); ++j )
        if (VS(j)->size()!=np) return 0;
    return np;
}

//**************************************************************************************************
//*  class COlc - interactive initialization                                 
//**************************************************************************************************

//! Prompt for curve description, and initialize things.

void COlc::askCurve( const string& quest )
{
 ask_again:
    string resp = sask( quest+" (or e(c|i)(r|w); h for help) ?" );
    if        ( resp=="h" ) {
        cout <<
            "curve definition:\n"
            "- either an arithmetic expression,\n"
            "    containing the running variable t,\n"
            "    for instance: p0+p1*t\n"
            "- or external evaluation mode e(c|i)(r|w)\n"
            "       next question will ask for external program\n"
            "    c: program reads pars from command line\n"
            "    i: program reads #pars, pars from standard input\n"
            "    r: program reads #points, points from standard input\n"
            "    w: program writes default grid, Frida interpolates\n";
        goto ask_again;
    } else if ( resp.size()==3 && resp[0]=='e' &&
                ( resp[1]=='c' || resp[1]=='i' ) &&
                ( resp[2]=='r' || resp[2]=='w' ) ) {
        if      ( resp[1]=='c' )
            scrInpMode = COlc::_CMDLIN;
        else if ( resp[1]=='i' )
            scrInpMode = COlc::_STDIN;
        else
            throw "BUG: impossible case";
        if      ( resp[2]=='r' )
            evaMode = COlc::_SCR;
        else if ( resp[2]=='w' )
            evaMode = COlc::_SCR_INTP;
        else
            throw "BUG: impossible case";
        expr = sask( "Curve evaluation command ?" );
        nP = iask( "Number of parameters ?" );
    } else {
        parseFunction( resp );
    }
    if( nP>100 )
        throw "PLAUSIBILITY CHECK FAILED: unexpected number of parameters np = "
            + S(nP);
}

//**************************************************************************************************
//*  class COlc - low-level initialization                                   
//**************************************************************************************************

//! Set auxiliary parameters to default values.

void COlc::curveDefaults()
{
    for( int ip=0; ip<nP; ip++ )
        PCo.push_back( CCoord("p" + S(ip), "") );
    weighing = COlc::_VAR;
}


//! Parse function expression.

void COlc::parseFunction( const string& _expr )
{
    evaMode = COlc::_EXPR;
    expr = _expr;
    T = user_xaxparse( expr.c_str() );
    if( !T->has_dummy() )
        cerr << "WARNING: curve has no formal argument t\n";
    nP = T->npar();
}

//**************************************************************************************************
//*  class COlc - function evaluation                                        
//**************************************************************************************************

//! Vectorial evaluation of a curve given by an expression tree.

void COlc::curve_val_vec_expr( vector<double>* v, vector<double>* dv,
                               const vector<double>& vt, int k, int j ) const
{
    CResult val;
    CContext ctx( k, j );
    ctx.want_error = dv;
    if ( dv && dv->size()!=v->size() )
        throw "BUG: inconsistent call of curve_val_vec_expr";
    if        ( !T->has_dummy() ) {  // curve does not depend on t
        T->tree_val( val, ctx );
        if ( val.vectorial )
            throw "BUG: tree_curve_val did not get scalar result";
        for( int i=0; i<v->size(); ++i ) {
            (*v)[i] = val.r;
            if( dv )
                (*dv)[i] = val.dr;
        }
    } else {
        ctx.request_VT( &vt );
        T->tree_val( val, ctx );
        if ( !val.vectorial )
            throw "BUG: tree_curve_val did not get vectorial result";
        *v = val.v;
        if( dv )
            *dv = val.dv;
    }
}

//! Vectorial evaluation of a curve given by a system call.

void COlc::curve_val_vec_syst( vector<double>* v, vector<double>* dv,
                               const vector<double>& vt, int k, int j ) const
{
    if( dv )
        throw "NOT YET IMPLEMENTED: curve_val_vec_syst with error bar";
    
    // Create FIFO to receive return values.
          // TODO: choose name as: triv::next_tmp_file( "/tmp/frida.%d" );
    string ret_fifonam = string("/tmp/ret-") + getenv( "LOGNAME" );
    triv::system( "rm -f "+ret_fifonam );
    if ( mkfifo( ret_fifonam.c_str(), 0666 ) )
        throw "SYSTEM ERROR cannot make " + ret_fifonam;
    
    string cmd = expr;
    string inp_fifo = "";
    string errmsg;
    string par_fifonam;
    
    // Create FIFO for passing pars or/and grid via stdin.
    if ( scrInpMode==_STDIN || evaMode == _SCR ) {
        par_fifonam = string("/tmp/par-") + getenv( "LOGNAME" );
        triv::system( "rm -f "+par_fifonam );
        if ( mkfifo( par_fifonam.c_str(), 0666 ) )
            throw "SYSTEM ERROR cannot make " + par_fifonam;
        inp_fifo = " < " + par_fifonam;
    }
    // Parameters passed on command line.
    if ( scrInpMode==_CMDLIN ){
        string pars = "";
        for ( int ip=0; ip<nP; ++ip )
            pars += " " + str( format("%g ") % VC(j)->P[ip] );
        cmd += pars;
    }

    // Start the external process.
    cmd += inp_fifo;
    cmd += " > " + ret_fifonam + " &";
    errmsg = "command '" + cmd + "'";
    system( cmd.c_str() );

    // Feed parameters and x values to FIFO.
    if( inp_fifo!="" ){
        int par_fifo;
        // we use open instead of fopen or ofstream,
        // because we need non-blocking mode.
        if ( !( par_fifo = open( par_fifonam.c_str(), O_WRONLY ) ) )
            throw "SYSTEM ERROR cannot open " + par_fifonam;
        fcntl( par_fifo, F_SETFL, O_NONBLOCK);

        if ( scrInpMode==_STDIN ){
            string out = str( format("%i ") % nP );
            write( par_fifo, out.c_str(), out.size() );
            for ( int ip=0; ip<nP; ++ip ) {
                string out = str( format("%16.8g ") % VC(j)->P[ip] );
                write( par_fifo, out.c_str(), out.size() );
            }
        }
        if ( evaMode==_SCR ) {
            string out = str( format("%i ") % vt.size() );
            write( par_fifo, out.c_str(), out.size() );
            for ( int i=0; i<vt.size(); ++i ) {
                string out = str( format("%16.8g ") % vt[i] );
                write( par_fifo, out.c_str(), out.size() );
            }
        }
        close( par_fifo );
    }
    //cout << "DEB cvv: #vt=" << vt.size() << ", called " << errmsg << "\n";

    ifstream FF( ret_fifonam );
    if( !FF.is_open() )
        throw "BUG: cannot read results of system call from " + ret_fifonam;

    if        ( evaMode==_SCR ) {
        for ( int i=0; i<vt.size(); ++i ) {
            string line;
            if( !getline( FF, line ) )
                throw "BUG: failed to read line " + S(i) +
                    " from " + ret_fifonam;
            istringstream iss(line);
            double x;
            try {
                iss >> x >> (*v)[i];
            } catch( exception &ex ) {
                throw S(ex.what()) + ": returns no x y pair for t["+S(i)+"]="+
                    S(vt[i])+" but '"+line+"'";
            }
            if( fabs(x-vt[i])>1e-4*fabs(vt[i]+x) )
                throw errmsg + " returns x=" + S(x) + " instead of " +
                    S(vt[i]) + " in line '"+line+"'";
            if ( !isfinite((*v)[i]) )
                throw "infinite y value in line '"+line+"'";
        }
    } else if ( evaMode==_SCR_INTP ) {
        CSpec S;
        double x, y;
        string line;
        while( getline( FF, line ) ) {
            istringstream iss(line);
            try {
                iss >> x >> y;
            } catch( exception &ex ) {
                throw S(ex.what()) + ": invalid line '"+line+"'";
            }
            S.push_xy( x, y );
        }
        // interpolation:
        if ( S.size()<2 )
            throw errmsg + " returns only " + S(S.size()) + " valid lines";
        //cout << "DEB cvv: intpol "<<vt.front()<<" .. "<<vt.back()<<"\n";
        S.intpol( vt, v, NULL );
    } else
        throw "BUG: invalid evaMode";
    FF.close();
}

//! Evaluate a curve, vectorial version. Used in fitting and in plotting.

void COlc::curve_val_vec( vector<double>* v, vector<double>* dv,
                          const vector<double>& vt, int k, int j ) const
  // input: the function arguments vt[], and the context k, j
  // output: the curve evaluation result *ret[]
{
    int n = vt.size();
    if( !n )
        throw "BUG: vectorial curve evaluation requested with empty t vector";
    v->resize( n );
    if( dv )
        dv->resize( n );

    if ( evaMode==_EXPR )
        curve_val_vec_expr( v, dv, vt, k, j );
    else if ( evaMode==_SCR || evaMode==_SCR_INTP )
        curve_val_vec_syst( v, dv, vt, k, j );
    else
        throw "BUG: unexpected evaluation mode";
}

//! Dito, scalar version, to obtain isolated points.

void COlc::curve_val_sca( double *r, double *dr, double farg, int k, int j ) const
{
    vector<double> vt(1, farg), v(1);
    if( dr ){
        vector<double> dv(1);
        curve_val_vec( &v, &dv, vt, k, j );
        *dr = dv[0];
    } else {
        curve_val_vec( &v, NULL, vt, k, j );
    }
    *r = v[0];
}

//**************************************************************************************************
//*  class COlc - inspect                                                    
//**************************************************************************************************

//! Returns text lines to describe file-wide settings.

vector<string> COlc::infoFile() const
{
    vector<string> ret;
    string out = "";
    if      ( evaMode==_SCR )
        out = "script: ";
    else if ( evaMode==_SCR_INTP )
        out = "script (followed by interpolation): ";
    ret.push_back( out + expr );
    string txt = "";
    if( kd!=-1 )
        txt += "data file: " + S(kd) + ", ";
    if( kconv!=-1 )
        txt += "conv file: " + S(kconv) + ", ";
    txt += "weighing: " + weight_str();
    ret.push_back( txt );
    return ret;
}

//! Returns header for scan parameter list.

string COlc::infoScanHeader() const
{
    const int PINFO_WRD_SIZ=15;
    char wrd[PINFO_WRD_SIZ];
    string lin;
    // z's and parameters:
    lin = "j   ";
    for ( int iz=0; iz<nZ(); iz++ ) {
        snprintf( wrd, PINFO_WRD_SIZ, "z%i           ", iz );
        lin += wrd;
    }
    for ( int ip=0; ip<nP; ip++ ) {
        snprintf( wrd, PINFO_WRD_SIZ, "p%d            ", ip );
        lin += wrd;
    }
    lin += "fit chi^2         1-R^2";
    return lin;
}

//! Returns a table of j,z,p lines.

string COlc::pInfoCat() const
{
    vector<string> vs = infoFile();
    vs.push_back( infoScanHeader() );
    for( int j=0; j<nJ(); j++ )
        vs.push_back( infoLine( j ) );
    string ret = "";
    for( int i=0; i<vs.size(); ++i )
        ret += vs[i] + "\n";
    return ret;
}

//! Returns description of fit weight.

string COlc::weight_str() const
{
    switch ( weighing ) {
    case _LIN:
        return "constant";
    case _LOG:
        return "logarithmic";
    case _VAR:
        return "with reciprocal variance (data and curve)";
    case _VARC:
        return "with reciprocal variance (curve only)";
    case _VARD:
        return "with reciprocal variance (data only)";
    }
    throw "BUG: unexpected weight";
}
